/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * @file remotecontrol.c
 *
 * Copyright (c) 2011 Robert Meyer, Frank Meyer - frank(at)fli4l.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */

#include <inttypes.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>

#include "irmpconfig.h"
#include "irmp.h"
#include "irsndconfig.h"
#include "irsnd.h"

/*-----------------------------------------------------------------------------------------------------------------------
 * LEDs
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define LED_1_PORT          PORTD
#define LED_1_DDR           DDRD
#define LED_1               6

#define LED_2_PORT          PORTD
#define LED_2_DDR           DDRD
#define LED_2               7

#define LED_3_PORT          PORTB
#define LED_3_DDR           DDRB
#define LED_3               0

#define LED_1_ON            LED_1_PORT |= 1<<LED_1
#define LED_1_OFF           LED_1_PORT &= ~(1<<LED_1)
#define LED_2_ON            LED_2_PORT |= 1<<LED_2
#define LED_2_OFF           LED_2_PORT &= ~(1<<LED_2)
#define LED_3_ON            LED_3_PORT |= 1<<LED_3
#define LED_3_OFF           LED_3_PORT &= ~(1<<LED_3)

/*-----------------------------------------------------------------------------------------------------------------------
 * Initialize LEDs
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
led_init (void)
{
    LED_1_PORT &= ~(1<<LED_1);
    LED_1_DDR |= (1<<LED_1);

    LED_2_PORT &= ~(1<<LED_2);
    LED_2_DDR |= (1<<LED_2);

    LED_3_PORT &= ~(1<<LED_3);
    LED_3_DDR |= (1<<LED_3);
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Switch LEDs on
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
leds_on (void)
{
    LED_1_ON;
    LED_2_ON;
    LED_3_ON;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Switch LEDs off
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
leds_off (void)
{
    LED_1_OFF;
    LED_2_OFF;
    LED_3_OFF;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Keys
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define KEY_POWER_PORT      PORTC
#define KEY_POWER_DDR       DDRC
#define KEY_POWER_PIN       PINC
#define KEY_POWER           4
#define KEY_POWER_VALUE     0

#define KEY_UP_PORT         PORTC
#define KEY_UP_DDR          DDRC
#define KEY_UP_PIN          PINC
#define KEY_UP              0
#define KEY_UP_VALUE        1

#define KEY_DOWN_PORT       PORTC
#define KEY_DOWN_DDR        DDRC
#define KEY_DOWN_PIN        PINC
#define KEY_DOWN            1
#define KEY_DOWN_VALUE      2

#define KEY_LEFT_PORT       PORTC
#define KEY_LEFT_DDR        DDRC
#define KEY_LEFT_PIN        PINC
#define KEY_LEFT            2
#define KEY_LEFT_VALUE      3

#define KEY_RIGHT_PORT      PORTC
#define KEY_RIGHT_DDR       DDRC
#define KEY_RIGHT_PIN       PINC
#define KEY_RIGHT           3
#define KEY_RIGHT_VALUE     4

#define MAX_KEYS            5
#define MAX_DEVICES         3

#define KEY_SELECT_PORT     PORTC
#define KEY_SELECT_DDR      DDRC
#define KEY_SELECT_PIN      PINC
#define KEY_SELECT          5
#define KEY_SELECT_VALUE    5

#define KEY_PROG_VALUE      6

static volatile uint8_t global_key = 0xFF;

/*-----------------------------------------------------------------------------------------------------------------------
 * Initialize keys
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
key_init (void)
{
    KEY_SELECT_DDR &= ~(1<<KEY_SELECT);
    KEY_SELECT_PORT |= (1<<KEY_SELECT);

    KEY_POWER_DDR &= ~(1<<KEY_POWER);
    KEY_POWER_PORT |= (1<<KEY_POWER);

    KEY_UP_DDR &= ~(1<<KEY_UP);
    KEY_UP_PORT |= (1<<KEY_UP);

    KEY_DOWN_DDR &= ~(1<<KEY_DOWN);
    KEY_DOWN_PORT |= (1<<KEY_DOWN);

    KEY_LEFT_DDR &= ~(1<<KEY_LEFT);
    KEY_LEFT_PORT |= (1<<KEY_LEFT);

    KEY_RIGHT_DDR &= ~(1<<KEY_RIGHT);
    KEY_RIGHT_PORT |= (1<<KEY_RIGHT);

    PCMSK1 |= (1<<PCINT8) | (1<<PCINT9) | (1<<PCINT10) | (1<<PCINT11) | (1<<PCINT12) | (1<<PCINT13);
    PCICR |= (1<<PCIE1);
}

/*-----------------------------------------------------------------------------------------------------------------------
 * poll keys
 *-----------------------------------------------------------------------------------------------------------------------
 */
static uint8_t
key_poll (void)
{
    uint8_t value;

    if (!(KEY_SELECT_PIN & (1<<KEY_SELECT)))
    {
        value = KEY_SELECT_VALUE;
    }
    else if (!(KEY_POWER_PIN & (1<<KEY_POWER)))
    {
        value = KEY_POWER_VALUE;
    }
    else if (!(KEY_UP_PIN & (1<<KEY_UP)))
    {
        value = KEY_UP_VALUE;
    }
    else if (!(KEY_DOWN_PIN & (1<<KEY_DOWN)))
    {
        value = KEY_DOWN_VALUE;
    }
    else if (!(KEY_LEFT_PIN & (1<<KEY_LEFT)))
    {
        value = KEY_LEFT_VALUE;
    }
    else if (!(KEY_RIGHT_PIN & (1<<KEY_RIGHT)))
    {
        value = KEY_RIGHT_VALUE;
    }
    else
    {
        value = 0xFF;
    }

    return value;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * IROUT
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define IR_PORT             PORTB
#define IR_DDR              DDRB
#define IR_PIN              PINB
#define IR                  6

#define IROUT_PORT          PORTD
#define IROUT_DDR           DDRD
#define IROUT               3

/*-----------------------------------------------------------------------------------------------------------------------
 * Initialize IR receiver and sender
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
ir_init (void)
{
    IR_PORT &= ~(1<<IR);                                                        // disable pullup
    IR_DDR  &= ~(1<<IR);                                                        // set IR to input

    IROUT_PORT &= ~(1<<IROUT);                                                  // set IROUT to 0
    IROUT_DDR |= (1<<IROUT);                                                    // set IROUT to output

    irmp_init ();
    irsnd_init();
}

/*-----------------------------------------------------------------------------------------------------------------------
 * IR-REC
 *-----------------------------------------------------------------------------------------------------------------------
 */

#define REC_PORT            PORTB
#define REC_DDR             DDRB
#define REC                 1

/*-----------------------------------------------------------------------------------------------------------------------
 * Initialize IR receiver
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
rec_init (void)
{
    REC_PORT &= ~(1<<REC);                                                      // set output to 0 (disable receiver)
    REC_DDR  |= (1<<REC);                                                       // set pin to output

    irmp_init ();
    irsnd_init();
}


/*-----------------------------------------------------------------------------------------------------------------------
 * Switch IR receiver on
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
rec_on (void)
{
    REC_PORT |= (1<<REC);                                                       // enable receiver
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Switch IR receiver off
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
rec_off (void)
{
    REC_PORT &= ~(1<<REC);                                                      // disable receiver
}


/*-----------------------------------------------------------------------------------------------------------------------
 * Initialize timer
 *-----------------------------------------------------------------------------------------------------------------------
 */
void
timer_init (void)
{
  OCR1A   =  (F_CPU / F_INTERRUPTS) - 1;                                    // compare value: 1/15000 of CPU frequency
  TCCR1B  = (1 << WGM12) | (1 << CS10);                                     // switch CTC Mode on, set prescaler to 1

#if defined (__AVR_ATmega8__) || defined (__AVR_ATmega16__) || defined (__AVR_ATmega32__) || defined (__AVR_ATmega64__) || defined (__AVR_ATmega162__) 
  TIMSK  = 1 << OCIE1A;                                                     // OCIE1A: Interrupt by timer compare
#else
  TIMSK1  = 1 << OCIE1A;                                                    // OCIE1A: Interrupt by timer compare
#endif  // __AVR...
}

static uint16_t idle_cnt;

/*-----------------------------------------------------------------------------------------------------------------------
 * Timer interrupt: called 15000 times per second
 *-----------------------------------------------------------------------------------------------------------------------
 */
ISR(TIMER1_COMPA_vect)
{
    static uint16_t cnt;
    uint8_t         k;
    static uint8_t  last_k = 0xFF;

    if (! irsnd_ISR())
    {
        (void) irmp_ISR();                                                  // call irmp ISR
    }

    k = key_poll ();

    if (k == 0xFF && last_k != KEY_SELECT_VALUE)
    {
        cnt = 0;

        if (idle_cnt < 0xFFFF)
        {
            idle_cnt++;
        }
    }
    else
    {
        idle_cnt = 0;

        if (k == last_k)
        {
            if (cnt < 0xFFFF)
            {
                cnt++;
            }

            if (k < MAX_KEYS && cnt >= (F_INTERRUPTS / 15))                 // 1/15 sec
            {
                cnt = 0;
                last_k = 0xFF;
                global_key = k;
            }
            else if (k == KEY_SELECT_VALUE && cnt >= F_INTERRUPTS)          // 1 sec
            {
                cnt = 0;
                last_k = 0xFF;
                global_key = KEY_PROG_VALUE;
            }
        }
        else
        {
            if (last_k == KEY_SELECT_VALUE)
            {
                if (cnt >= (F_INTERRUPTS / 15))                             // 1/15 sec
                {
                    global_key = KEY_SELECT_VALUE;
                    k = 0xFF;
                }
            }

            cnt = 0;
            last_k = k;
        }
    }
}

/*-----------------------------------------------------------------------------------------------------------------------
 * get a key
 *-----------------------------------------------------------------------------------------------------------------------
 */
static uint8_t
getkey (void)
{
    uint8_t k;
    cli();
    k = global_key;
    global_key = 0xFF;
    sei();
    return k;
}

static uint8_t     device_number = 0;

/*-----------------------------------------------------------------------------------------------------------------------
 * Show active device
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
show_device_number (uint8_t device_number)
{
    leds_off();

    switch (device_number)
    {
        case 0: LED_1_ON; break;
        case 1: LED_2_ON; break;
        case 2: LED_3_ON; break;
    }
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Called (back) from IRSND module
 *-----------------------------------------------------------------------------------------------------------------------
 */
void
led_callback (uint8_t on)
{
    if (on)
    {
        switch (device_number)
        {
            case 0: LED_1_ON; break;
            case 1: LED_2_ON; break;
            case 2: LED_3_ON; break;
        }
    }
    else
    {
        switch (device_number)
        {
            case 0: LED_1_OFF; break;
            case 1: LED_2_OFF; break;
            case 2: LED_3_OFF; break;
        }
    }
}

/*-----------------------------------------------------------------------------------------------------------------------
 * EEPROM routines
 *-----------------------------------------------------------------------------------------------------------------------
 */
static uint8_t     ee_protocol[MAX_DEVICES][MAX_KEYS]  EEMEM;
static uint16_t    ee_address[MAX_DEVICES][MAX_KEYS]   EEMEM;
static uint16_t    ee_command[MAX_DEVICES][MAX_KEYS]   EEMEM;

/*-----------------------------------------------------------------------------------------------------------------------
 * Read a byte from EEPROM
 *-----------------------------------------------------------------------------------------------------------------------
 */
static uint8_t
my_eeprom_read_byte (uint8_t * addr)
{
    uint8_t value;
    uint8_t tmp_sreg;

    tmp_sreg = SREG;
    cli();

    eeprom_busy_wait ();
    value = eeprom_read_byte (addr);

    SREG = tmp_sreg;
    return value;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Store a byte into EEPROM
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
my_eeprom_write_byte (uint8_t * addr, uint8_t value)
{
    uint8_t tmp_sreg;

    tmp_sreg = SREG;
    cli();

    eeprom_busy_wait ();
    eeprom_write_byte (addr, value);

    SREG = tmp_sreg;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Read a word from EEPROM
 *-----------------------------------------------------------------------------------------------------------------------
 */
static uint16_t
my_eeprom_read_word (uint16_t * addr)
{
    uint16_t value;
    uint8_t tmp_sreg;

    tmp_sreg = SREG;
    cli();

    eeprom_busy_wait ();
    value = eeprom_read_word (addr);

    SREG = tmp_sreg;
    return value;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Store a word into EEPROM
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
my_eeprom_write_word (uint16_t * addr, uint16_t value)
{
    uint8_t tmp_sreg;

    tmp_sreg = SREG;
    cli();

    eeprom_busy_wait ();
    eeprom_write_word (addr, value);

    SREG = tmp_sreg;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * MAIN
 *-----------------------------------------------------------------------------------------------------------------------
 */
int
main (void)
{
    IRMP_DATA   irmp_data;
    uint8_t     protocol[MAX_DEVICES][MAX_KEYS];
    uint16_t    address[MAX_DEVICES][MAX_KEYS];
    uint16_t    command[MAX_DEVICES][MAX_KEYS];
    uint8_t     cancelled;
    uint8_t     leds_are_on = FALSE;
    uint8_t     key;
    uint8_t     i;
    uint8_t     j;

    for (i = 0; i < MAX_DEVICES; i++)                                       // read IRMP code from EEPROM into RAM
    {
        for (j = 0; j < MAX_KEYS; j++)
        {
            protocol[i][j] = my_eeprom_read_byte (&ee_protocol[i][j]);
            address[i][j] = my_eeprom_read_word (&ee_address[i][j]);
            command[i][j] = my_eeprom_read_word (&ee_command[i][j]);
        }
    }

    led_init ();                                                            // initialize LED, KEY, IR, timer, receiver
    key_init ();
    ir_init ();
    rec_init ();
    timer_init ();
    irsnd_callback_ptr = led_callback;

    sei ();                                                                 // enable interrupts

    set_sleep_mode (SLEEP_MODE_PWR_DOWN);                                   // set CPU into powerdown mode
    sleep_enable ();
    sleep_cpu();

    while (1)
    {
        if ((key = getkey()) != 0xFF)                                       // read key, pressed?
        {                                                                   // yes
            if (key == KEY_SELECT_VALUE)                                    // select key pressed?
            {
                if (leds_are_on)                                            // leds already on, switch to next device
                {
                    device_number++;

                    if (device_number >= MAX_DEVICES)
                    {
                        device_number = 0;
                    }
                }

                show_device_number (device_number);                         // show current device
                leds_are_on = TRUE;
            }
            else if (key == KEY_PROG_VALUE)                                 // switch into learning mode
            {
                while (irmp_get_data (&irmp_data))                          // empty irmp data
                {
                    ;
                }

                leds_on();                                                  // switch all LEDs on: we are ready to receive IR signal
                rec_on ();                                                  // enable receiver

                while (getkey() != 0xFF)                                    // empty key data
                {
                    ;
                }

                cancelled = FALSE;

                while (! irmp_get_data (&irmp_data))                        // receive IR signal
                {
                    if (getkey() < MAX_KEYS)                                // pressed a key?
                    {                                                       // yes
                        cancelled = TRUE;                                   // cancel learning mode
                        break;
                    }
                }

                leds_off();                                                 // let LEDs blink
                _delay_ms (250);
                leds_on();
                _delay_ms (250);

                if (! cancelled)
                {
                    while ((key = getkey()) >= MAX_KEYS)                    // get key to store IR code
                    {
                        if (key == KEY_SELECT_VALUE)
                        {
                            break;
                        }
                    }

                    if (key < MAX_KEYS)                                     // valid key pressed?
                    {
                        protocol[device_number][key] = irmp_data.protocol;  // store IR code into RAM and EEPROM
                        address[device_number][key] = irmp_data.address;
                        command[device_number][key] = irmp_data.command;

                        my_eeprom_write_byte (&ee_protocol[device_number][key], protocol[device_number][key]);
                        my_eeprom_write_word (&ee_address[device_number][key], address[device_number][key]);
                        my_eeprom_write_word (&ee_command[device_number][key], command[device_number][key]);

                        leds_on();                                          // let LEDs blink 2 times: code stored
                        _delay_ms (250);
                        leds_off();
                        _delay_ms (250);
                        leds_on();
                        _delay_ms (250);
                        leds_off();
                    }
                }

                rec_off ();                                                 // disable receiver to save energy (current)
            }
            else if (key < MAX_KEYS && protocol[device_number][key] != 0xFF)
            {                                                               // send IR code (stored in RAM)
                irmp_data.flags = 0;                                        // reset flags
                irmp_data.protocol = protocol[device_number][key];          // set protocol
                irmp_data.address = address[device_number][key];            // set address
                irmp_data.command = command[device_number][key];            // set command
                irsnd_send_data (&irmp_data, 0);                            // send IR code now
            }
        }
        else
        {
            if (idle_cnt >= F_INTERRUPTS / 2)                               // 1/2 second idle...
            {
                cli ();                                                     // reset idle counter
                idle_cnt = 0;
                sei ();

                leds_are_on = FALSE;                                        // reset LED flag
                leds_off ();                                                // switch LEDs off
                sleep_cpu ();                                               // get into powerdown mode
            }
        }
    }
}

ISR(PCINT1_vect)                                                            // key pressed (PCINT service routine)
{
    return;                                                                 // intentionally left blank ;-)
}
